/*
 *  linux/drivers/fusion/mpi.c
 *      High performance SCSI / Fibre Channel device driver.
 *      For use with PCI adapter(s):
 *          LSI SYM409xx (Fibre Channel)
 *      running LSI Logic Fusion MPT (Message Passing Technology) firmware.
 *
 *  Credits:
 *      This driver would not exist if not for Alan Cox's development
 *      of the linux i2o driver.
 *      A huge debt of gratitude is owed to David S. Miller (DaveM)
 *      for fixing much of the stupid and broken stuff in the early
 *      driver while porting to sparc64 platform.  THANK YOU!
 *
 *  Copyright (c) 1999-2000 LSI Logic Corporation
 *  Originally By: Steven J. Ralston
 *  (mailto:Steve.Ralston@lsil.com)
 *
 *  $Id: mpi.c,v 1.70 2000/09/29 23:31:06 nromer Exp $
 */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    NO WARRANTY
    THE PROGRAM IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OR
    CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED INCLUDING, WITHOUT
    LIMITATION, ANY WARRANTIES OR CONDITIONS OF TITLE, NON-INFRINGEMENT,
    MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. Each Recipient is
    solely responsible for determining the appropriateness of using and
    distributing the Program and assumes all risks associated with its
    exercise of rights under this Agreement, including but not limited to
    the risks and costs of program errors, damage to or loss of data,
    programs or equipment, and unavailability or interruption of operations.

    DISCLAIMER OF LIABILITY
    NEITHER RECIPIENT NOR ANY CONTRIBUTORS SHALL HAVE ANY LIABILITY FOR ANY
    DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING WITHOUT LIMITATION LOST PROFITS), HOWEVER CAUSED AND
    ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
    USE OR DISTRIBUTION OF THE PROGRAM OR THE EXERCISE OF ANY RIGHTS GRANTED
    HEREUNDER, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include <linux/config.h>
#include <linux/version.h>

#if defined(MODULE) && !defined(__GENKSYMS__)
#include <linux/module.h>
#endif

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/malloc.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <asm/io.h>
#ifdef CONFIG_MTRR
#include <asm/mtrr.h>
#endif /* CONFIG_MTRR */

#include <linux/kdev_t.h>
#include <linux/blkdev.h>

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#include "mpi.h"

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
#define my_NAME     "Fusion MPT base driver"
#define my_VERSION  LINUX_MPI_VERSION_COMMON

#ifdef MODULE
MODULE_AUTHOR(MoDuLeAuThOr);
MODULE_DESCRIPTION(my_NAME);

/*
 *  cmd line parameters
 */
MODULE_PARM(PortIo, "0-1i");
MODULE_PARM_DESC(PortIo, "[0]=Use mmap, 1=Use port io");
MODULE_PARM(HardReset, "0-1i");
MODULE_PARM_DESC(HardReset, "0=Disable HardReset, [1]=Enable HardReset");
#endif
static int PortIo = 0;
static int HardReset = 1;

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Public data...
 */
ASCQ_Table_t *ASCQ_TablePtr = NULL;
int ASCQ_TableSz = 0;
const char **ScsiOpcodesPtr = NULL;
int LanCIdx = 0;

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Private data...
 */

STATIC MPI_ADAPTER_TRACKER   MpiAdapters;
STATIC MPI_CALLBACK          MpiCallbacks[myMPI_MAX_CALLBACKS];
                                     /* Lookup table */
STATIC MPI_ADAPTER          *mpi_adapters[myMPI_MAX_ADAPTERS] = {0};
#ifdef MPI_DEBUG
static int max_replies_serviced = 0;
#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Forward protos...
 */
STATIC void  mpi_interrupt(int irq, void *bus_id, struct pt_regs *r);

STATIC  int __init mpi_adapter_install(struct pci_dev *dev);
STATIC void        mpi_adapter_dispose(MPI_ADAPTER *ioc);

STATIC U32   GetIocState(MPI_ADAPTER *ioc, int cooked);
STATIC int   GetIocFacts(MPI_ADAPTER *ioc);
STATIC int   GetPortFacts(MPI_ADAPTER *ioc);
STATIC int   SendIocInit(MPI_ADAPTER *ioc);
STATIC int   SendPortEnable(MPI_ADAPTER *ioc);
STATIC int   KickStart(MPI_ADAPTER *ioc);
STATIC int   SendIocReset(MPI_ADAPTER *ioc, U8 reset_type);
STATIC int   PrimeIocFifos(MPI_ADAPTER *ioc);
STATIC int   HandShakeReqAndReply(MPI_ADAPTER *ioc, int reqBytes, U32 *req, int replyBytes, U16 *u16reply);
STATIC int   WaitForDoorbellAck(MPI_ADAPTER *ioc);
STATIC int   WaitForDoorbellInt(MPI_ADAPTER *ioc);
STATIC int   WaitForDoorbellReply(MPI_ADAPTER *ioc);
STATIC int   GetLanConfigPages(MPI_ADAPTER *ioc);

#ifdef CONFIG_PROC_FS
STATIC int  procmpi_create              (void);
STATIC int  procmpi_destroy             (void);
STATIC int  procmpi_read_info           (char *page, char **start, off_t off,
                                         int count, int *eof, void *data);
/*STATIC int  procmpi_info              (char *buf, char **start, off_t offset, int len);*/
/*STATIC void procmpi_print_ioc_facts   (MPI_ADAPTER *ioc, char *buf, int *size, int len);*/
STATIC struct proc_dir_entry *procmpi_root_dir = NULL;
#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/* 20000207 -sralston
 *  GRRRRR...  IOSpace (port i/o) register access (for the 909) is back!
 * 20000517 -sralston
 *  Let's trying going back to default mmap register access...
 */

STATIC inline U32 CHIPREG_READ32(volatile U32 *a)
{
	if (PortIo)
		return inl((unsigned long)a);
	else
		return readl(a);
}

STATIC inline void CHIPREG_WRITE32(volatile U32 *a, U32 v)
{
	if (PortIo)
		outl(v, (unsigned long)a);
	else
		writel(v, a);
}


/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_interrupt - MPI adapter (IOC) specific interrupt handler
 *	@irq: irq number (not used)
 *	@bus_id: bus identifier cookie == pointer to MPI_ADAPTER structure
 *	@r: pt_regs pointer (not used)
 *
 *	This routine is registered via the request_irq() kernel API call,
 *	and handles all interrupts generated from a specific MPI adapter
 *	(also referred to as a IO Controller or IOC).
 *	This routine must clear the interrupt from the adapter and does
 *	so by reading the reply FIFO.  Multiple replies may be processed
 *	per single call to this routine; up to myMPI_MAX_REPLIES_PER_ISR
 *	which is currently set to 10 in mpi.h.
 *
 *	This routine handles register-level access of the adapter but
 *	dispatches (calls) a protocol-specific callback routine to handle
 *	the protocol-specific details of the MPI request completion.
 */
STATIC void
mpi_interrupt(int irq, void *bus_id, struct pt_regs *r)
{
	MPI_ADAPTER	*ioc;
	MPI_FRAME_HDR	*mf;
	MPI_FRAME_HDR	*mr = NULL;
	U32		 pa;
	U32		*m;
	U32		 intstat;
	int		 req_idx;
	int		 cb_idx;
	int		 freeme;
	int		 lanturbo = 0;
	int		 count = 0;
	int		 frames_processed = 0;

	ioc = bus_id;

	/* get raw IntStatus reg value */
	intstat = CHIPREG_READ32(&ioc->chip->IntStatus);
	dirqprintk((KERN_INFO "mpi: %s:  !!! InTeRrUpT !!!\n", ioc->name));

	/*
	 *  Process reply and then doorbell interrupts, in that order.
	 */
	do {
		dirqprintk((KERN_INFO "mpi: read #%d of IntStatus=%08x\n",
			    count+1, intstat));
		/*
		 *  Check reply interrupt.  If set, drain the reply FIFO!
		 *  (I've seen up to 10 replies processed in this loop, so far...)
		 * Update: I've seen up to 9182 replies processed in this loop! ??
		 */
		if (intstat & MPI_HIS_REPLY_MESSAGE_INTERRUPT)
		{
		while (    (frames_processed < myMPI_MAX_REPLIES_PER_ISR)
		        && ((pa = CHIPREG_READ32(&ioc->chip->ReplyFifo)) != 0xFFFFFFFF)
		      )
		{
			mf = NULL;
			cb_idx = 0;
			lanturbo = 0;

			/*  Check for non-TURBO reply!  */
			if (pa & MPI_ADDRESS_REPLY_A_BIT) {
				dma_addr_t reply_dma_addr;

				/* non-TURBO reply!  Uh-oh - something's up...
				 *
				 * YIKES!  Newest turbo reply mechanism; get address
				 *  via left shift 1 (get rid of MPI_ADDRESS_REPLY_A_BIT)!
				 */
				reply_dma_addr = (pa = (pa << 1));

				/* Map DMA address of reply header to cpu address. */
				m = (U32 *)
					((U8 *)ioc->reply_frames +
					 (reply_dma_addr - ioc->reply_frames_dma));

#if defined(MPI_DEBUG) || defined(MPI_DEBUG_MSG_FRAME)
				printk(KERN_WARNING "mpi: Got non-TURBO reply=%p\n", m);
				{
					int  i, n = (le32_to_cpu(m[0]) & 0x00FF0000) >> 16;

					printk(KERN_WARNING " ");
					for (i=0; i<n; i++)
						printk(" %08x", le32_to_cpu(m[i]));
					printk("\n");
				}
#endif
				mr = (MPI_FRAME_HDR *) m;
				req_idx = le16_to_cpu(mr->u.frame.hwhdr.msgctxu.fld.req_idx);
				cb_idx = mr->u.frame.hwhdr.msgctxu.fld.cb_idx;
			} else {
				dirqprintk((KERN_INFO "mpi: Got TURBO reply(=%08x)\n", pa));
				if ((pa >> MPI_CONTEXT_REPLY_TYPE_SHIFT) == MPI_CONTEXT_REPLY_TYPE_LAN) {
					mr = (MPI_FRAME_HDR *) ((void*)0 + pa);
					if ((pa & 0x58000000) == 0x58000000)
						req_idx = pa & 0x0000FFFF;
					else
						req_idx = -42;
					lanturbo = 1;
					cb_idx = LanCIdx;
				} else {
					req_idx = pa & 0x0000FFFF;
					cb_idx = (pa & 0x00FF0000) >> 16;
				}
				pa = 0;					/* No reply flush! */
			}

			/*  Calculate ptr to original request!  */
			if (req_idx != -42)
				mf = MPI_INDEX_2_MFPTR(ioc, req_idx);

			/*  Check for IO callback! */
			freeme = 0;
			if (MpiCallbacks[cb_idx] != NULL) {
				/*  Do the callback!  */
				if (lanturbo)
					freeme = (*(MpiCallbacks[cb_idx]))(ioc, NULL, mr);
				else
					freeme = (*(MpiCallbacks[cb_idx]))(ioc, mf, mr);
			}

			if (pa) {
				/*  Flush (non-TURBO) reply with a WRITE!  */
				CHIPREG_WRITE32(&ioc->chip->ReplyFifo, pa);
			}

			if (freeme) {
				unsigned long flags;

				/*  Put Request back on FreeQ!  */
				spin_lock_irqsave(&ioc->FreeQlock, flags);
				Q_ADD_TAIL(&ioc->FreeQ, &mf->u.frame.linkage, MPI_FRAME_HDR);
				spin_unlock_irqrestore(&ioc->FreeQlock, flags);
			}

			frames_processed++;
			dirqprintk((KERN_INFO "mpi: ISR processed frame=%d\n", frames_processed));
		}
		}	/* if reply interrupt set, drain reply FIFO */

#ifdef MPI_DEBUG
		if (frames_processed > max_replies_serviced) {
			max_replies_serviced = frames_processed;
			dprintk((KERN_INFO "mpi: MaxRepliesServiced=%d (per single ISR)\n",
					       max_replies_serviced));
		}
#endif

		if (!(intstat & MPI_HIS_REPLY_MESSAGE_INTERRUPT)) {
			printk(KERN_INFO "mpi: %s: Hmmm... IntStatus=%08x (cnt=%d)\n",
			                 ioc->name, intstat, count);
			/*  Clear IntStatus reg  */
			CHIPREG_WRITE32(&ioc->chip->IntStatus, 0);
		}

		if (frames_processed >= myMPI_MAX_REPLIES_PER_ISR) {
			dirqprintk((KERN_INFO "mpi: %s: ISR Processed %d replies.",
				    ioc->name, frames_processed));
			dirqprintk(("  Giving this ISR a break!\n"));
			/* Hmmm...
			 *  Maybe this is a problem?  20000722 -sralston
			 *  Removing the following line seems to eliminate ABORT
			 *  (due to lost IO?) which leads to hard hang described
			 *  above.
			 * CHIPREG_WRITE32(&ioc->chip->IntStatus, 0);
			 */
			return;
		}

		mb();
		intstat = CHIPREG_READ32(&ioc->chip->IntStatus);
		count++;

	} while (intstat);
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_register - Register protocol-specific callback handler
 *	@cb_func: callback function
 *
 *	This routine is called by the SCSI or LAN protocol driver so that
 *	it can register it's reply callback routine.
 *	NOTES: The SCSI protocol driver currently calls this routine twice
 *	in order to register separate callbacks; one for "normal" SCSI IO
 *	and another for MpiScsiTaskMgmt requests.
 *
 *	Returns integer valued protocol callback "handle" (in 7-0 range).
 */
int
mpi_register(MPI_CALLBACK cbfunc)
{
	int r = -1;
	int i;

	for (i = (myMPI_MAX_CALLBACKS - 1); i; i--) {
		if (MpiCallbacks[i] == NULL) {
			MpiCallbacks[i] = cbfunc;
			r = i;
			MOD_INC_USE_COUNT;
			break;
		}
	}

	return r;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_deregister - Deregister protocol-specific callback handler
 *	@cb_idx: previously registered callback handle
 *
 *	Each protocol-specific driver should call this routine when it's
 *	module is unloaded.
 */
void
mpi_deregister(int cb_idx)
{
	if (cb_idx < myMPI_MAX_CALLBACKS) {
		MpiCallbacks[cb_idx] = NULL;
		MOD_DEC_USE_COUNT;
	}
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_register_ascqops_strings - Register SCSI ASC/ASCQ and SCSI
 *	OpCode strings from the isense driver
 *	@ascqTable: Pointer to ASCQ_Table_t structure
 *	@ascqtbl_sz: Number of entries in ASCQ_Table
 *	@opsTable: Pointer to array of SCSI OpCode strings (char pointers)
 *
 *	Specialized driver registration routine for the isense driver.
 */
int
mpi_register_ascqops_strings(ASCQ_Table_t *ascqTable, int ascqtbl_sz, const char **opsTable)
{
	int r = 0;

	if (ascqTable && ascqtbl_sz && opsTable) {
		ASCQ_TablePtr = ascqTable;
		ASCQ_TableSz = ascqtbl_sz;
		ScsiOpcodesPtr = opsTable;
		printk(KERN_INFO "mpi: English readable SCSI-3 strings enabled:-)\n");
		r = 1;
	}
	MOD_INC_USE_COUNT;

	return r;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_deregister_ascqops_strings - Deregister SCSI ASC/ASCQ and SCSI
 *			OpCode strings from the isense driver
 *
 *	Specialized driver deregistration routine for the isense driver.
 */
void
mpi_deregister_ascqops_strings(void)
{
	ASCQ_TablePtr = NULL;
	ASCQ_TableSz = 0;
	ScsiOpcodesPtr = NULL;
	printk(KERN_INFO "mpi: English readable SCSI-3 strings disabled)-:\n");
	MOD_DEC_USE_COUNT;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	MpiGetMsgFrame - Obtain a MPI request frame from the pool (of 1024)
 *			allocated per MPI adapter
 *	@handle: Handle of registered MPI protocol driver
 *	@iocid: IOC unique identifier (integer)
 *
 *	Returns pointer to a MPI request frame or %NULL if none are available.
 */
MPI_FRAME_HDR*
MpiGetMsgFrame(int handle, int iocid)
{
	MPI_FRAME_HDR *mf = NULL;
	MPI_ADAPTER *iocp;

	/* validate handle and ioc identifier */
	iocp = mpi_adapters[iocid];
	if (MpiCallbacks[handle] != NULL && iocp != NULL) {
		spin_lock(&iocp->FreeQlock);
		if (! Q_IS_EMPTY(&iocp->FreeQ)) {
			int req_offset;

			mf = iocp->FreeQ.head;
			Q_DEL_ITEM(&mf->u.frame.linkage);
			mf->u.frame.hwhdr.msgctxu.fld.cb_idx = handle;	/* byte */
			req_offset = (U8 *)mf - (U8 *)iocp->req_frames;
									/* U16! */
			mf->u.frame.hwhdr.msgctxu.fld.req_idx =
			            cpu_to_le16(req_offset / iocp->req_sz);
			mf->u.frame.hwhdr.msgctxu.fld.rsvd = 0;
		}
		spin_unlock(&iocp->FreeQlock);
	}
#ifdef MPI_DEBUG_MSG_FRAME
	printk(KERN_INFO "mpi: %s: MpiGetMsgFrame(%d,%d), got mf=%p\n",
	                 iocp->name, handle, iocid, mf);
#endif
	return mf;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	MpiPutMsgFrame - Send a protocol specific MPI request frame to a IOC
 *	@handle: Handle of registered MPI protocol driver
 *	@iocid: IOC unique identifier (integer)
 *	@mf: Pointer to MPI request frame
 *
 *	This routine posts a MPI request frame to the request post FIFO of a
 *	specific MPI adapter.
 */
void
MpiPutMsgFrame(int handle, int iocid, MPI_FRAME_HDR *mf)
{
	MPI_ADAPTER *iocp;

	iocp = mpi_adapters[iocid];
	if (iocp != NULL) {
		dma_addr_t mf_dma_addr;
		int req_offset;

		/* ensure values are reset properly! */
		mf->u.frame.hwhdr.msgctxu.fld.cb_idx = handle;		/* byte */
		req_offset = (U8 *)mf - (U8 *)iocp->req_frames;
									/* U16! */
		mf->u.frame.hwhdr.msgctxu.fld.req_idx = cpu_to_le16(req_offset / iocp->req_sz);
		mf->u.frame.hwhdr.msgctxu.fld.rsvd = 0;

#ifdef MPI_DEBUG_MSG_FRAME
		{
			U32 *m = mf->u.frame.hwhdr.__hdr;
			int  i, n = iocp->req_sz/4;

			printk(KERN_INFO "mpi: About to Put msg frame @ %p:\n", m);
			printk(KERN_INFO " ");
			for (i=0; i<n; i++) {
				if (i && ((i%8)==0))
					printk("\n" KERN_INFO " ");
				printk(" %08x", le32_to_cpu(m[i]));
			}
			printk("\n");
		}
#endif

		mf_dma_addr = iocp->req_frames_dma + req_offset;
		CHIPREG_WRITE32(&iocp->chip->RequestFifo, mf_dma_addr);
	}
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	MpiFreeMsgFrame - Place MPI request frame back on FreeQ
 *	@handle: Handle of registered MPI protocol driver
 *	@iocid: IOC unique identifier (integer)
 *	@mf: Pointer to MPI request frame
 *
 *	This routine places a MPI request frame back on the MPI adapter's
 *	FreeQ.
 */
void
MpiFreeMsgFrame(int handle, int iocid, MPI_FRAME_HDR *mf)
{
	MPI_ADAPTER *iocp;
	unsigned long flags;

	iocp = mpi_adapters[iocid];
	if (iocp != NULL) {
		/*  Put Request back on FreeQ!  */
		spin_lock_irqsave(&iocp->FreeQlock, flags);
		Q_ADD_TAIL(&iocp->FreeQ, &mf->u.frame.linkage, MPI_FRAME_HDR);
		spin_unlock_irqrestore(&iocp->FreeQlock, flags);
	}
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	SendHandShakeRequest - Send MPI request via doorbell handshake method
 *	@handle: Handle of registered MPI protocol driver
 *	@iocid: IOC unique identifier (integer)
 *	@reqBytes: Size of the request in bytes
 *	@req: Pointer to MPI request frame
 *
 *	This routine is used exclusively by mpiscsi to send MpiScsiTaskMgmt
 *	requests since they are required to be sent via doorbell handshake.
 *
 *	NOTE: It is the callers responsibility to byte-swap fields in the
 *	request which are greater than 1 byte in size.
 *
 *	Returns 0 for success, non-zero for failure.
 */
int
SendHandShakeRequest(int handle, int iocid, int reqBytes, U32 *req)
{
	MPI_ADAPTER	*iocp;
	int		 r = 0;

	iocp = mpi_adapters[iocid];
	if (iocp != NULL) {
		U8 *req_as_bytes;
		int i;

		/*
		 *  Emulate what MpiPutMsgFrame() does /wrt to sanity
		 *  setting cb_idx/req_idx.  But ONLY if this request
		 *  is in proper (pre-alloc'd) request buffer range...
		 */
		i = MFPTR_2_MPI_INDEX(iocp,(MPI_FRAME_HDR*)req);
		if (reqBytes >= 12 && i >= 0 && i < iocp->req_depth) {
			MPI_FRAME_HDR *mf = (MPI_FRAME_HDR*)req;
			mf->u.frame.hwhdr.msgctxu.fld.req_idx = cpu_to_le16(i);
			mf->u.frame.hwhdr.msgctxu.fld.cb_idx = handle;
		}

// YIKES!		/*  Make sure there are no doorbells  */
		CHIPREG_WRITE32(&iocp->chip->IntStatus, 0);

		CHIPREG_WRITE32(&iocp->chip->Doorbell,
				((MPI_FUNCTION_HANDSHAKE<<MPI_DOORBELL_FUNCTION_SHIFT) |
				 ((reqBytes/4)<<MPI_DOORBELL_ADD_DWORDS_SHIFT)));

		/*  Wait for IOC doorbell int  */
		if ((i = WaitForDoorbellInt(iocp)) < 0)
			return i;

		dhsprintk((KERN_INFO "mpi: %s: SendHandShakeRequest start, WaitCnt=%d\n",
				     iocp->name, i));

		CHIPREG_WRITE32(&iocp->chip->IntStatus, 0);

		if ((r = WaitForDoorbellAck(iocp)) < 0)
			return -2;

		/*  Send request via doorbell handshake  */
		req_as_bytes = (U8 *) req;
		for (i = 0; i < reqBytes/4; i++) {
			U32 word;

			word = ((req_as_bytes[(i*4) + 0] <<  0) |
				(req_as_bytes[(i*4) + 1] <<  8) |
				(req_as_bytes[(i*4) + 2] << 16) |
				(req_as_bytes[(i*4) + 3] << 24));
			CHIPREG_WRITE32(&iocp->chip->Doorbell, word);
			if ((r = WaitForDoorbellAck(iocp)) < 0) {
				r = -3;
				break;
			}
		}

// NEW!		// FIXME? Wait for Doorbell flag to drop?

		/*  Make sure there are no doorbells  */
		CHIPREG_WRITE32(&iocp->chip->IntStatus, 0);

	}

	return r;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_adapter_find_first - Find first MPI adapter pointer
 *
 *	Returns first MPI adapter pointer or %NULL if no MPI adapters
 *	are present.
 */
MPI_ADAPTER *
mpi_adapter_find_first(void)
{
	MPI_ADAPTER *this = NULL;

	if (! Q_IS_EMPTY(&MpiAdapters))
		this = MpiAdapters.head;

	return this;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 * 	mpi_adapter_find_next - Find next MPI adapter pointer
 * 	@prev: Pointer to previous MPI adapter
 *
 *	Returns next MPI adapter pointer or %NULL if there are no more.
 */
MPI_ADAPTER *
mpi_adapter_find_next(MPI_ADAPTER *prev)
{
	MPI_ADAPTER *next = NULL;

	if (prev && (prev->forw != (MPI_ADAPTER*)&MpiAdapters.head))
		next = prev->forw;

	return next;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_pci_scan - Scan PCI devices for MPI adapters
 *
 *	Returns count of MPI adapters found, keying off of PCI vendor and
 *	device_id's.
 */
int __init
mpi_pci_scan(void)
{
	struct pci_dev *dev;
	int count = 0;
	int r;

	dprintk((KERN_INFO "mpi: Checking for MPI adapters...\n"));

	/*
	 *  NOTE: The 929 (I believe) will appear as 2 separate PCI devices,
	 *        one for each channel.
	 */
	pci_for_each_dev(dev) {
		if (dev->vendor != 0x1000)
			continue;

		/* FIXME!
		 *  Also handle C1030 explicitly...
		 */
		if ((dev->device != 0x0621) &&
		    (dev->device != 0x0622) &&
#if 0
		    /* FIXME! C1030 */
		    (dev->device != 0x0xxx) &&
#endif
		    1) {
			dprintk((KERN_INFO "mpi: Skipping LSI device=%04xh\n", dev->device));
			continue;
		}

		dprintk((KERN_INFO "mpi: MPI adapter found: PCI bus/dfn=%02x/%02xh, class=%08x, id=%xh\n",
			 dev->bus->number, dev->devfn, dev->class, dev->device));

		if ((r = mpi_adapter_install(dev)) == 0)
			count++;
		else
			return r;
	}

	if (count == 0)
		return -ENODEV;

#ifdef CONFIG_PROC_FS
	if (procmpi_create() != 0)
		printk(KERN_WARNING "mpi: WARNING! /proc/mpi creation failed!\n");
#endif

	dprintk((KERN_INFO "mpi: %d MPI adapter channel%s found and installed.\n",
		 count, (count==1) ? "" : "s"));

	return count;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_verify_adapter - Given a unique IOC identifier, set pointer to
 *			the associated MPI adapter structure.
 *	@iocid: IOC unique identifier (integer)
 *	@iocpp: Pointer to pointer to IOC adapter
 *
 *	Returns iocid and sets iocpp.
 */
int
mpi_verify_adapter(int iocid, MPI_ADAPTER **iocpp)
{
	MPI_ADAPTER *p;

	*iocpp = NULL;
	if (iocid >= myMPI_MAX_ADAPTERS)
		return -1;

	p = mpi_adapters[iocid];
	if (p == NULL)
		return -1;

	*iocpp = p;
	return iocid;
}

#if 0
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
MPI_ADAPTER *
mpi_verify_device(int iocport, int btl)
{
	MPI_ADAPTER *this = NULL;
	Scsi_Device *SDpnt;

	if (Q_IS_EMPTY(&MpiAdapters))
		return this;

	this = MpiAdapters.head;
	do {
		U8 ioc, port, bus, target, lun;

		ioc = (iocport & 0xFF00) >> 8;
		port = (iocport & 0x00FF);
		bus = (btl & 0x00FF0000) >> 16;
		target = (btl & 0x0000FF00) >> 8;
		lun = (btl & 0x000000FF);

		if (this->sh != NULL) {
			SDpnt = this->sh->host_queue;
			dprintk(KERN_INFO "mpi: Searching for target=%d, lun=%d...\n", target, lun);

			while (SDpnt != NULL) {
				dprintk((KERN_INFO "mpi: SDpnt=%p, target=%d, lun=%d\n",
					 SDpnt, SDpnt->id, SDpnt->lun));
				if ((SDpnt->id == target) && (SDpnt->lun == lun))
					return this;

				SDpnt = SDpnt->next;
			}
		}

		this = this->forw;
	} while (this && (this != (MPI_ADAPTER*)&MpiAdapters.head));

	return NULL;
}
#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_adapter_install - Install a PCI intelligent MPI adapter
 *	@dev: Pointer to pci_dev structure
 *
 *	This routine performs all the steps necessary to bring the IOC of
 *	a MPI adapter to a OPERATIONAL state.  This includes registering
 *	memory regions, registering the interrupt, and allocating request
 *	and reply memory pools.
 *
 *	This routine also pre-fetches the LAN MAC address of a Fibre Channel
 *	MPI adapter.
 *
 *	Returns 0 for success, non-zero for failure.
 *
 *	TODO: Add support for polled controllers
 */
STATIC int __init
mpi_adapter_install(struct pci_dev *dev)
{
	MPI_ADAPTER	*ioc;
	char		*myname;
	U8		*mem;
	unsigned long	 mem_phys;
	unsigned long	 port;
	U32		 msize;
	U32		 psize;
	U32		 ioc_state;
	int		 i;
	int 		 cntdn;
	int		 len;
	int		 statefault = 0;

	ioc = kmalloc(sizeof(MPI_ADAPTER), GFP_KERNEL);
	if (ioc == NULL) {
		printk(KERN_ERR "mpi: ERROR: Insufficient memory to add adapter!\n");
		return -ENOMEM;
	}
	memset(ioc, 0, sizeof(*ioc));
	ioc->alloc_total = sizeof(MPI_ADAPTER);

	ioc->pcidev = dev;

	/*  Find lookup slot.  GRRRR...  */
	for (i=0; i < myMPI_MAX_ADAPTERS; i++) {
		if (mpi_adapters[i] == NULL) {
			ioc->id = i;		/* Assign adapter unique id (lookup) */
			break;
		}
	}
	if (i == myMPI_MAX_ADAPTERS) {
		printk(KERN_ERR "mpi: ERROR! mpi_adapters[%d] table overflow!\n", i);
		kfree(ioc);
		return -ENFILE;
	}

	mem_phys = msize = 0;
	port = psize = 0;
	for (i=0; i < DEVICE_COUNT_RESOURCE; i++) {
		if (dev->PCI_BASEADDR_FLAGS(i) & PCI_BASE_ADDRESS_SPACE) {
			/*  Get I/O space!  */
			port = dev->PCI_BASEADDR_START(i);
			psize = PCI_BASEADDR_SIZE(dev,i);
		} else {
			/*  Get memmap  */
			mem_phys = dev->PCI_BASEADDR_START(i);
			msize = PCI_BASEADDR_SIZE(dev,i);
			break;
		}
	}
	ioc->mem_size = msize;

	if (i == DEVICE_COUNT_RESOURCE) {
		printk(KERN_ERR "mpi: ERROR: MPI adapter has no memory regions defined!\n");
		kfree(ioc);
		return -EINVAL;
	}

	dprintk((KERN_INFO "mpi: MPI adapter @ %lx, msize=%dd bytes\n", mem_phys, msize));
	dprintk((KERN_INFO "mpi: (also, port i/o @ %lx, psize=%dd bytes)\n", port, psize));
	dprintk((KERN_INFO "mpi: Using %s register access method\n", PortIo ? "PortIo" : "MemMap"));

	mem = NULL;
	if (! PortIo) {
		/*  Get logical ptr for PciMem0 space  */
		/*mem = ioremap(mem_phys, msize);*/
		mem = ioremap(mem_phys, 0x100);
		if (mem == NULL) {
			printk(KERN_ERR "mpi: ERROR: Unable to map adapter memory.\n");
			kfree(ioc);
			return -EINVAL;
		}
		ioc->memmap = mem;
	}
	dprintk((KERN_INFO "mpi: mem = %p, mem_phys = %lx\n", mem, mem_phys));

	if (PortIo) {
		U8 *pmem = (U8*)port;
		ioc->mem_phys = port;
		ioc->chip = (SYSIF_REGS*)pmem;
	} else {
		ioc->mem_phys = mem_phys;
		ioc->chip = (SYSIF_REGS*)mem;
	}

	ioc->chip_type = FCUNK;
	if (dev->device == 0x0621)
		ioc->chip_type = FC909;
	else if (dev->device == 0x0622)
		ioc->chip_type = FC929;
	//else if (dev->device == 0x????)
	//	ioc->chip_type = C1030;

	myname = "iocN";
	len = strlen(myname);
	memcpy(ioc->name, myname, len+1);
	ioc->name[len-1] = '0' + ioc->id;

	Q_INIT(&ioc->FreeQ, MPI_FRAME_HDR);
	spin_lock_init(&ioc->FreeQlock);

	/*  Disable all!  */
	CHIPREG_WRITE32(&ioc->chip->IntMask, 0xFFFFFFFF);
	ioc->active = 0;

	ioc->pci_irq = -1;
	if (dev->irq) {
		i = request_irq(dev->irq, mpi_interrupt, SA_SHIRQ, ioc->name, ioc);

		if (i < 0) {
			printk(KERN_ERR "mpi: ERROR: %s: unable to allocate interrupt %d\n",
					ioc->name, dev->irq);
			iounmap(mem);
			kfree(ioc);
			return -EBUSY;
		}

		ioc->pci_irq = dev->irq;

		pci_set_master(dev);			/* ?? */

		dprintk((KERN_INFO "mpi: INFO: %s installed at interrupt %d\n", ioc->name, dev->irq));
	}

	/*  tack onto tail of our MPI adapter list  */
	Q_ADD_TAIL(&MpiAdapters, ioc, MPI_ADAPTER);

	/* Set lookup ptr. */
	mpi_adapters[ioc->id] = ioc;


	/*  Enable! (reply interrupt)  */
	CHIPREG_WRITE32(&ioc->chip->IntMask, ~(SYSIF_INTMASK_REPLYINT));
	ioc->active = 1;

	/*  Get current [raw] IOC state  */
	ioc_state = GetIocState(ioc, 0);
	dhsprintk((KERN_INFO "mpi: %s initial [raw] state=%08x\n", ioc->name, ioc_state));

	/*
	 *  Check to see if IOC got left/stuck in doorbell handshake
	 *  grip of death.  If so, hard reset the IOC.
	 */
	if (ioc_state & MPI_DOORBELL_ACTIVE) {
		statefault = 1;
		printk(KERN_WARNING "mpi: %s: Uh-oh, unexpected doorbell active!\n",
		       ioc->name);
	}

	/* NEW!  20000817 -sralston
	 *  Check to see if IOC is in FAULT state.
	 *  If so, hard reset the IOC.
	 */
	if ((ioc_state & MPI_IOC_STATE_MASK) == MPI_IOC_STATE_FAULT) {
		statefault = 2;
		printk(KERN_WARNING "mpi: %s: Uh-oh, IOC is in FAULT state!!!\n",
		       ioc->name);
		printk(KERN_WARNING "           FAULT code = %04xh\n",
		       ioc_state & MPI_DOORBELL_DATA_MASK);
	}

	if (statefault) {
		(void) KickStart(ioc);
	}

	/*
	 *  Loop here waiting for IOC to come READY.
	 */
	i = 0;
	cntdn = 10 * 100;
	while ((ioc_state = GetIocState(ioc, 1)) != MPI_IOC_STATE_READY) {
		if (ioc_state == MPI_IOC_STATE_OPERATIONAL) {
			/*
			 *  BIOS or previous driver load left IOC in OP state.
			 *  Reset messaging FIFOs.
			 */
			dprintk((KERN_WARNING "mpi: %s: Sending IOC msg unit reset!\n",
			         ioc->name));
			(void) SendIocReset(ioc, MPI_FUNCTION_IOC_MESSAGE_UNIT_RESET);
		} else if (ioc_state == MPI_IOC_STATE_RESET) {
			/*
			 *  Something is wrong.  Try to get IOC back
			 *  to a known state.
			 */
			dprintk((KERN_WARNING "mpi: %s: Sending IO unit reset!\n", ioc->name));
			(void) SendIocReset(ioc, MPI_FUNCTION_IO_UNIT_RESET);
		}

		i++; cntdn--;
		if (!cntdn) {
			printk(KERN_ERR "mpi: ERROR: %s: Wait IOC_READY state timeout! (cnt=%d)\n",
			       ioc->name, i);
			free_irq(ioc->pci_irq, ioc);
			iounmap(mem);
			kfree(ioc);
			return -ETIME;
		}

		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(1);
	}

	if (statefault) {
		printk(KERN_WARNING "mpi: %s: Whew!  Recovered from %s\n",
		       ioc->name, statefault==1 ? "stuck handshake" : "IOC FAULT");
	}

	/*  Get IOC facts! (first time, ioc->facts0 and ioc->pfacts0)  */
	if ((i = GetIocFacts(ioc)) != 0)
		return i;

	if ((i = SendIocInit(ioc)) != 0)
		return i;

	/*
	 *  Prime reply & request queues!
	 *  (mucho alloc's)
	 */
	if ((i = PrimeIocFifos(ioc)) != 0)
		return i;

	/*  Get IOC facts again! (2nd time, ioc->factsN and ioc->pfactsN)  */
	if ((i = GetIocFacts(ioc)) != 0)
		return i;

	if (ioc->pfacts0.ProtocolFlags & MPI_PORTFACTS_PROTOCOL_LAN) {
		/*
		 *  Pre-fetch the ports LAN MAC address!
		 *  (LANPage1_t stuff)
		 */
		(void) GetLanConfigPages(ioc);
#ifdef MPI_DEBUG
		{
			U8 *a = (U8*)&ioc->lan_cnfg_page1.HardwareAddressLow;
			dprintk((KERN_INFO "mpi: LanAddr = 0x(%02x:%02x:%02x:%02x:%02x:%02x)\n",
					a[5], a[4], a[3], a[2], a[1], a[0] ));
		}
#endif
	}

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	mpi_adapter_dispose - Free all resources associated with a MPI adapter
 *	@this: Pointer to MPI adapter structure
 *
 *	This routine unregisters h/w resources and frees all alloc'd memory
 *	associated with a MPI adapter structure.
 */
STATIC void
mpi_adapter_dispose(MPI_ADAPTER *this)
{
	if (this != NULL) {
		int sz, sz_first, sz_last;

		sz_first = this->alloc_total;

		if (this->pci_irq > 0) {
			/*  Disable all!  */
			CHIPREG_WRITE32(&this->chip->IntMask, 0xFFFFFFFF);
			this->active = 0;
			free_irq(this->pci_irq, this);
		}

		if (this->memmap != NULL)
			iounmap((U8 *) this->memmap);

#if defined(CONFIG_MTRR) && 0
		if (this->mtrr_reg > 0) {
			mtrr_del(this->mtrr_reg, 0, 0);
			dprintk((KERN_INFO "mpi: MTRR region de-registered\n"));
		}
#endif

		if (this->reply_alloc != NULL) {
			sz = (this->reply_sz * this->reply_depth) + 128;
			pci_free_consistent(this->pcidev, sz,
					    this->reply_alloc, this->reply_alloc_dma);
			this->reply_frames = NULL;
			this->reply_alloc = NULL;
			this->alloc_total -= sz;
		}

		if (this->req_alloc != NULL) {
			sz = (this->req_sz * this->req_depth) + 128;
			/*
			 *  Rounding UP to nearest 4-kB boundary here...
			 */
			sz = ((sz + 0x1000UL - 1UL) / 0x1000) * 0x1000;
			pci_free_consistent(this->pcidev, sz,
					    this->req_alloc, this->req_alloc_dma);
			this->req_frames = NULL;
			this->req_alloc = NULL;
			this->alloc_total -= sz;
		}

		if (this->sense_buf_pool != NULL) {
			sz = (this->req_depth * 256);
			pci_free_consistent(this->pcidev, sz,
					    this->sense_buf_pool, this->sense_buf_pool_dma);
			this->sense_buf_pool = NULL;
			this->alloc_total -= sz;
		}

		/*  Zap the adapter lookup ptr!  */
		mpi_adapters[this->id] = NULL;

		sz_last = this->alloc_total;
		dprintk((KERN_INFO "mpi: free'd %d of %d bytes\n",
		         sz_first-sz_last+(int)sizeof(*this), sz_first));
		kfree(this);
	}
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	GetIocState - Get the current state of a MPI adapter
 *	@ioc: Pointer to MPI_ADAPTER structure
 *	@cooked: Request raw or cooked IOC state
 *
 *	Returns all IOC Doorbell register bits if cooked==0, else just the
 *	Doorbell bits in MPI_IOC_STATE_MASK.
 */
STATIC U32
GetIocState(MPI_ADAPTER *ioc, int cooked)
{
	U32 s, sc;

	/*  Get!  */
	s = CHIPREG_READ32(&ioc->chip->Doorbell);
	dprintk((KERN_INFO "mpi: %s: raw state = %08x\n", ioc->name, s));
	sc = s & MPI_IOC_STATE_MASK;

	/*  Save!  */
	ioc->last_state = sc;

	return cooked ? sc : s;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	GetIocFacts - Send IOCFacts request to MPI adapter
 *	@ioc: Pointer to MPI_ADAPTER structure
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
GetIocFacts(MPI_ADAPTER *ioc)
{
	IOCFacts_t		 get_facts;
	IOCFactsReply_t		*facts;
	int			 r;
	int			 req_sz;
	int			 reply_sz;

	/*  IOC *must* NOT be in RESET state!  */
	if (ioc->last_state == MPI_IOC_STATE_RESET) {
		printk(KERN_ERR "mpi: ERROR: Can't get IOCFacts - %s NOT READY! (%08x)\n",
				ioc->name,
				ioc->last_state );
		return -4;
	}

	facts = &ioc->facts0;
	/*  Nth (2,3,...) time thru? (been here, done that?)  */
	if (ioc->facts0.Function == MPI_FUNCTION_IOC_FACTS) {
		facts = &ioc->factsN;
	}

	/*  Destination (reply area)...  */
	reply_sz = sizeof(*facts);
	memset(facts, 0, reply_sz);

	/*  Request area (get_facts on the stack right now!)  */
	req_sz = sizeof(get_facts);
	memset(&get_facts, 0, req_sz);

	get_facts.Function = MPI_FUNCTION_IOC_FACTS;
	/*  Assert: All other get_facts fields are zero!  */

	dprintk((KERN_INFO "mpi: Sending IocFacts%s request\n",
			   facts == &ioc->facts0 ? "0" : "N" ));

	/* No non-zero fields in the get_facts request are greater than
	 * 1 byte in size, so we can just fire it off as is.
	 */
	r = HandShakeReqAndReply(ioc,
			req_sz, (U32*)&get_facts,
			reply_sz, (U16*)facts);
	if (r != 0)
		return r;

	/* Now byte swap the necessary fields before inspecting the
	 * contents of the reply.
	 */
	facts->MsgVersion = le16_to_cpu(facts->MsgVersion);
	facts->MsgContext = le32_to_cpu(facts->MsgContext);
	facts->IOCStatus = le16_to_cpu(facts->IOCStatus);
	facts->IOCLogInfo = le32_to_cpu(facts->IOCLogInfo);
	facts->ReplyQueueDepth = le16_to_cpu(facts->ReplyQueueDepth);
	facts->RequestFrameSize = le16_to_cpu(facts->RequestFrameSize);
	facts->FWVersion = le16_to_cpu(facts->FWVersion);
	facts->ProductID = le16_to_cpu(facts->ProductID);
	facts->CurrentHostMfaHighAddr =
		le32_to_cpu(facts->CurrentHostMfaHighAddr);
	facts->GlobalCredits = le16_to_cpu(facts->GlobalCredits);
	facts->CurrentSenseBufferHighAddr =
		le32_to_cpu(facts->CurrentSenseBufferHighAddr);
	facts->CurReplyFrameSize =
		le16_to_cpu(facts->CurReplyFrameSize);

	/*  Set the REQUEST queue size & depth we'll use...  */

	ioc->req_sz = MIN(myMPI_REQ_SIZE, facts->RequestFrameSize * 4);
	ioc->req_depth = MIN(myMPI_REQ_DEPTH, facts->GlobalCredits);

	/*  Set the REPLY queue size & depth we'll use...  */

	/* BUG? FIX?  20000516 -nromer & sralston 
	 *  GRRR...  The following did not translate well from MPI v0.09:
	 *	ioc->reply_sz = MIN(myMPI_REPLY_SIZE, facts->ReplySize * 4);
	 *  to 0.10:
	 *    ioc->reply_sz = MIN(myMPI_REPLY_SIZE, facts->BlockSize * 4);
	 *  Was trying to minimally optimize to smallest possible reply size
	 *  (and greatly reduce kmalloc size).  But LAN may need larger reply?
	 *
	 *  So for now, just set reply size to request size.  FIXME?
	 */
	ioc->reply_sz = ioc->req_sz;
	ioc->reply_depth = MIN(myMPI_REPLY_DEPTH, facts->ReplyQueueDepth);

	dprintk((KERN_INFO "mpi: reply_sz=%3d, reply_depth=%4d\n",
			   ioc->reply_sz, ioc->reply_depth));
	dprintk((KERN_INFO "mpi: req_sz  =%3d, req_depth  =%4d\n",
			   ioc->req_sz, ioc->req_depth));

	/*  Get port facts!  */
	if ( (r = GetPortFacts(ioc)) != 0 )
		return r;

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	GetPortFacts - Send PortFacts request to MPI adapter
 *	@ioc: Pointer to MPI_ADAPTER structure
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
GetPortFacts(MPI_ADAPTER *ioc)
{
	PortFacts_t		 get_pfacts;
	PortFactsReply_t	*pfacts;
	int			 i;
	int			 req_sz;
	int			 reply_sz;

	/*  IOC *must* NOT be in RESET state!  */
	if (ioc->last_state == MPI_IOC_STATE_RESET) {
		printk(KERN_ERR "mpi: ERROR: Can't get PortFacts - %s NOT READY! (%08x)\n",
				ioc->name,
				ioc->last_state );
		return -4;
	}

	pfacts = &ioc->pfacts0;
	/*  Nth (2,3,...) time thru? (been here, done that?)  */
	if (ioc->pfacts0.Function == MPI_FUNCTION_PORT_FACTS) {
		pfacts = &ioc->pfactsN;
	}

	/*  Destination (reply area)...  */
	reply_sz = sizeof(*pfacts);
	memset(pfacts, 0, reply_sz);

	/*  Request area (get_pfacts on the stack right now!)  */
	req_sz = sizeof(get_pfacts);
	memset(&get_pfacts, 0, req_sz);

	get_pfacts.Function = MPI_FUNCTION_PORT_FACTS;
	/*  Assert: All other get_pfacts fields are zero!  */

	/* No non-zero fields in the get_pfacts request are greater than
	 * 1 byte in size, so we can just fire it off as is.
	 */
	i = HandShakeReqAndReply(ioc, req_sz, (U32*)&get_pfacts,
				reply_sz, (U16*)pfacts);
	if (i != 0)
		return i;

	/* Now byte swap the necessary fields in the response. */
	pfacts->MsgContext = le32_to_cpu(pfacts->MsgContext);
	pfacts->IOCStatus = le16_to_cpu(pfacts->IOCStatus);
	pfacts->IOCLogInfo = le32_to_cpu(pfacts->IOCLogInfo);
	pfacts->MaxDevices = le16_to_cpu(pfacts->MaxDevices);
	pfacts->PortSCSIID = le16_to_cpu(pfacts->PortSCSIID);
	pfacts->ProtocolFlags = le16_to_cpu(pfacts->ProtocolFlags);
	pfacts->MaxPostedCmdBuffers = le16_to_cpu(pfacts->MaxPostedCmdBuffers);
	pfacts->MaxPersistentIDs = le16_to_cpu(pfacts->MaxPersistentIDs);
	pfacts->MaxLanBuckets = le16_to_cpu(pfacts->MaxLanBuckets);

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	SendIocInit - Send IOCInit request to MPI adapter
 *	@ioc: Pointer to MPI_ADAPTER structure
 *
 *	Send IOCInit followed by PortEnable to bring IOC to OPERATIONAL state.
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
SendIocInit(MPI_ADAPTER *ioc)
{
	IOCInit_t		 ioc_init;
	MPIDefaultReply_t	 init_reply;
	U32			 state;
	int			 r;
	int			 count;
	int			 cntdn;

	memset(&ioc_init, 0, sizeof(ioc_init));
	memset(&init_reply, 0, sizeof(init_reply));

	ioc_init.WhoInit      = MPI_WHOINIT_HOST_DRIVER;
/*	ioc_init.ChainOffset  = 0;				*/
	ioc_init.Function     = MPI_FUNCTION_IOC_INIT;
/*	ioc_init.Flags        = 0;				*/

	/*ioc_init.MaxDevices   = 16;*/
	ioc_init.MaxDevices   = 255;
/*	ioc_init.MaxBuses     = 16;				*/
	ioc_init.MaxBuses     = 1;

/*	ioc_init.MsgFlags     = 0;				*/
/*	ioc_init.MsgContext   = cpu_to_le32(0x00000000);	*/
	ioc_init.ReplyFrameSize  = cpu_to_le16(ioc->reply_sz);	/* in BYTES */
	ioc_init.HostMfaHighAddr = cpu_to_le32(0);	/* Say we 32-bit! for now */

	dprintk((KERN_INFO "mpi: Sending IOCInit (req @ %p)\n", &ioc_init));

	r = HandShakeReqAndReply(ioc, sizeof(IOCInit_t), (U32*)&ioc_init,
			sizeof(MPIDefaultReply_t), (U16*)&init_reply);
	if (r != 0)
		return r;

	/* No need to byte swap the multibyte fields in the reply
	 * since we don't even look at it's contents.
	 */

	if ((r = SendPortEnable(ioc)) != 0)
		return r;

	/* YIKES!  SUPER IMPORTANT!!!
	 *  Poll IocState until _OPERATIONAL while IOC is doing
	 *  LoopInit and TargetDiscovery!
	 */
	count = 0;
	cntdn = 10 * 100;
	state = GetIocState(ioc, 1);
	while (state != MPI_IOC_STATE_OPERATIONAL && --cntdn) {
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(1);

		if (!cntdn) {
			printk(KERN_ERR "mpi: ERROR: %s: Wait IOC_OP state timeout! (cnt=%d)\n",
				       	ioc->name, count);
			return -9;
		}

		state = GetIocState(ioc, 1);
		count++;
	}
	dhsprintk((KERN_INFO "mpi: INFO: Wait IOC_OP state (cnt=%d)\n", count));

	return r;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	SendPortEnable - Send PortEnable request to MPI adapter
 *	@ioc: Pointer to MPI_ADAPTER structure
 *
 *	Send PortEnable to bring IOC to OPERATIONAL state.
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
SendPortEnable(MPI_ADAPTER *ioc)
{
	PortEnable_t		 port_enable;
	MPIDefaultReply_t	 reply_buf;
	int	 i;
	int	 req_sz;
	int	 reply_sz;

	/*  Destination...  */
	reply_sz = sizeof(MPIDefaultReply_t);
	memset(&reply_buf, 0, reply_sz);

	req_sz = sizeof(PortEnable_t);
	memset(&port_enable, 0, req_sz);

	port_enable.Function     = MPI_FUNCTION_PORT_ENABLE;
/*	port_enable.ChainOffset  = 0;				*/
/*	port_enable.MsgFlags     = 0;				*/
/*	port_enable.MsgContext   = cpu_to_le32(0x00000000);	*/

	dprintk((KERN_INFO "mpi: Sending PortEnable (req @ %p)\n", &port_enable));

	i = HandShakeReqAndReply(ioc, req_sz, (U32*)&port_enable,
			reply_sz, (U16*)&reply_buf);
	if (i != 0)
		return i;

	/* We do not even look at the reply, so we need not
	 * swap the multi-byte fields.
	 */

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	KickStart - Perform hard reset of MPI adapter
 *	@ioc: Pointer to MPI_ADAPTER structure
 *
 *	This routine places MPI adapter in diagnostic mode via the
 *	WriteSequence register, and then performs a hard reset of adapter
 *	via the Diagnostic register.
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
KickStart(MPI_ADAPTER *ioc)
{
	U32 diagval;
	U32 ioc_state;
	int cnt = 0;

	printk(KERN_WARNING "mpi: KickStarting %s!\n", ioc->name);

	if (!HardReset) {
		printk(KERN_WARNING "mpi: Oops, HardReset option DISABLED!\n");
		return -1;
	}

	/*  Use "Diagnostic reset" method! (only thing available!)  */

	if (ioc->chip_type == FC909 || ioc->chip_type == FC929) {
		/*
		 *  Extra read to handle 909 B.0 chip problem with reset
		 *  logic not finishing the RAM access before hard reset hits.
		 *  (? comment taken from NT SYMMPI source)
		 */
		(void) CHIPREG_READ32(&ioc->chip->Fubar);

		/*  Write magic sequence to WriteSequence register  */
		CHIPREG_WRITE32(&ioc->chip->WriteSequence, 0x0F);
		CHIPREG_WRITE32(&ioc->chip->WriteSequence, 0x04);
		CHIPREG_WRITE32(&ioc->chip->WriteSequence, 0x0B);
		CHIPREG_WRITE32(&ioc->chip->WriteSequence, 0x02);
		CHIPREG_WRITE32(&ioc->chip->WriteSequence, 0x07);
		CHIPREG_WRITE32(&ioc->chip->WriteSequence, 0x0D);

		/*  Now hit the reset bit in the Diagnostic register  */
		CHIPREG_WRITE32(&ioc->chip->Diagnostic, 0x04);

		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(10);

		if ((diagval = CHIPREG_READ32(&ioc->chip->Diagnostic)) & 0x42) {
			printk(KERN_ERR "mpi: %s: Diagnostic reset FAILED!\n",
			                ioc->name);
		}

		for (cnt=0; cnt<200; cnt++) {
			if ((ioc_state = GetIocState(ioc, 1)) == MPI_IOC_STATE_READY)
				break;
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(10);
		}

		if (cnt >= 200) {
			printk(KERN_ERR "mpi: %s: Failed to come READY after reset!\n",
			                ioc->name);
			return -2;
		}
	} else if (ioc->chip_type == C1030) {
		/* FIXME! */
	} else {
		return -3;
	}

	dprintk((KERN_INFO "mpi: %s: KickStart successful! (cnt=%d)\n",
	                   ioc->name, cnt));
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	SendIocReset - Send IOCReset request to MPI adapter
 *	@ioc: Pointer to MPI_ADAPTER structure
 *	@reset_type: reset type, expected values are
 *	%MPI_FUNCTION_IOC_MESSAGE_UNIT_RESET or %MPI_FUNCTION_IO_UNIT_RESET
 *
 *	Send IOCReset request to the MPI adapter.
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
SendIocReset(MPI_ADAPTER *ioc, U8 reset_type)
{
	int r;

	printk(KERN_WARNING "mpi: %s: Sending IOC reset(0x%02x)!\n",
	                    ioc->name, reset_type);
	CHIPREG_WRITE32(&ioc->chip->Doorbell, reset_type<<MPI_DOORBELL_FUNCTION_SHIFT);
	if ((r = WaitForDoorbellAck(ioc)) < 0)
		return r;
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	PrimeIocFifos - Initialize IOC request and reply FIFOs
 *	@ioc: Pointer to MPI_ADAPTER structure
 *
 *	This routine allocates memory for the MPI reply and request frame
 *	pools, and primes the IOC reply FIFO with reply frames.
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
PrimeIocFifos(MPI_ADAPTER *ioc)
{
	MPI_FRAME_HDR *mf;
	unsigned long b;
	dma_addr_t aligned_mem_dma;
	U8 *mem, *aligned_mem;
	int i, sz;

	/*  Prime reply FIFO...  */

	if (ioc->reply_frames == NULL) {
		sz = (ioc->reply_sz * ioc->reply_depth) + 128;
		mem = pci_alloc_consistent(ioc->pcidev, sz, &ioc->reply_alloc_dma);
		if (mem == NULL)
			goto out_fail;

		memset(mem, 0, sz);
		ioc->alloc_total += sz;
		ioc->reply_alloc = mem;
/*				   "_________0_________0___4		*/
		dprintk((KERN_INFO "mpi: %s.reply_alloc  @ %p[%08x], sz=%d bytes\n",
			 ioc->name, mem, ioc->reply_alloc_dma, sz));

		b = (unsigned long) mem;
		b = (b + (0x80UL - 1UL)) & ~(0x80UL - 1UL); /* round up to 128-byte boundary */
		aligned_mem = (U8 *) b;
		ioc->reply_frames = (MPI_FRAME_HDR *) aligned_mem;
		ioc->reply_frames_dma =
			(ioc->reply_alloc_dma + (aligned_mem - mem));
		aligned_mem_dma = ioc->reply_frames_dma;
		dprintk((KERN_INFO "mpi: %s.reply_frames @ %p[%08x]\n",
			 ioc->name, aligned_mem, aligned_mem_dma));

		for (i = 0; i < ioc->reply_depth; i++) {
			/*  Write each address to the IOC!  */
			CHIPREG_WRITE32(&ioc->chip->ReplyFifo, aligned_mem_dma);
			aligned_mem_dma += ioc->reply_sz;
		}
	}


	/*  Request FIFO - WE manage this!  */

	if (ioc->req_frames == NULL) {
		sz = (ioc->req_sz * ioc->req_depth) + 128;
		/*
		 *  Rounding UP to nearest 4-kB boundary here...
		 */
		sz = ((sz + 0x1000UL - 1UL) / 0x1000) * 0x1000;

		mem = pci_alloc_consistent(ioc->pcidev, sz, &ioc->req_alloc_dma);
		if (mem == NULL)
			goto out_fail;

		memset(mem, 0, sz);
		ioc->alloc_total += sz;
		ioc->req_alloc = mem;
		dprintk((KERN_INFO "mpi: %s.req_alloc    @ %p[%08x], sz=%d bytes\n",
			 ioc->name, mem, ioc->req_alloc_dma, sz));

		b = (unsigned long) mem;
		b = (b + (0x80UL - 1UL)) & ~(0x80UL - 1UL); /* round up to 128-byte boundary */
		aligned_mem = (U8 *) b;
		ioc->req_frames = (MPI_FRAME_HDR *) aligned_mem;
		ioc->req_frames_dma =
			(ioc->req_alloc_dma + (aligned_mem - mem));
		aligned_mem_dma = ioc->req_frames_dma;

		dprintk((KERN_INFO "mpi: %s.req_frames   @ %p[%08x]\n",
			 ioc->name, aligned_mem, aligned_mem_dma));

		for (i = 0; i < ioc->req_depth; i++) {
			mf = (MPI_FRAME_HDR *) aligned_mem;

			/*  Queue REQUESTs *internally*!  */
			Q_ADD_TAIL(&ioc->FreeQ.head, &mf->u.frame.linkage, MPI_FRAME_HDR);
			aligned_mem += ioc->req_sz;
		}

#if defined(CONFIG_MTRR) && 0
		/*
		 *  Enable Write Combining MTRR for IOC's memory region.
		 *  (at least as much as we can; "size and base must be
		 *  multiples of 4 kiB"
		 */
		ioc->mtrr_reg = mtrr_add(ioc->req_alloc_dma,
					 sz,
					 MTRR_TYPE_WRCOMB, 1);
		dprintk((KERN_INFO "mpi: MTRR region registered (base:size=%08x:%x)\n",
				ioc->req_alloc_dma,
				sz ));
#endif

	}

	if (ioc->sense_buf_pool == NULL) {
		sz = (ioc->req_depth * 256);
		ioc->sense_buf_pool = pci_alloc_consistent(ioc->pcidev,
							   sz,
							   &ioc->sense_buf_pool_dma);
		if (ioc->sense_buf_pool == NULL)
			goto out_fail;

		ioc->alloc_total += sz;
	}

	return 0;

out_fail:
	if (ioc->reply_alloc != NULL) {
		sz = (ioc->reply_sz * ioc->reply_depth) + 128;
		pci_free_consistent(ioc->pcidev,
				    sz,
				    ioc->reply_alloc, ioc->reply_alloc_dma);
		ioc->reply_frames = NULL;
		ioc->reply_alloc = NULL;
		ioc->alloc_total -= sz;
	}
	if (ioc->req_alloc != NULL) {
		sz = (ioc->req_sz * ioc->req_depth) + 128;
		/*
		 *  Rounding UP to nearest 4-kB boundary here...
		 */
		sz = ((sz + 0x1000UL - 1UL) / 0x1000) * 0x1000;
		pci_free_consistent(ioc->pcidev,
				    sz,
				    ioc->req_alloc, ioc->req_alloc_dma);
#if defined(CONFIG_MTRR) && 0
		if (ioc->mtrr_reg > 0) {
			mtrr_del(ioc->mtrr_reg, 0, 0);
			dprintk((KERN_INFO "mpi: MTRR region de-registered\n"));
		}
#endif
		ioc->req_frames = NULL;
		ioc->req_alloc = NULL;
		ioc->alloc_total -= sz;
	}
	if (ioc->sense_buf_pool != NULL) {
		sz = (ioc->req_depth * 256);
		pci_free_consistent(ioc->pcidev,
				    sz,
				    ioc->sense_buf_pool, ioc->sense_buf_pool_dma);
		ioc->sense_buf_pool = NULL;
	}
	return -1;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	HandShakeReqAndReply - Send MPI request to and receive reply from IOC
 *			via doorbell handshake method
 *	@ioc: Pointer to MPI_ADAPTER structure
 *	@reqBytes: Size of the request in bytes
 *	@req: Pointer to MPI request frame
 *	@replyBytes: Expected size of the reply in bytes
 *	@u16reply: Pointer to area where reply should be written
 *
 *	NOTE: It is the callers responsibility to byte-swap fields in the
 *	request which are greater than 1 byte in size.  It is also the
 *	callers responsibility to byte-swap response fields which are
 *	greater than 1 byte in size.
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
HandShakeReqAndReply(MPI_ADAPTER *ioc, int reqBytes, U32 *req, int replyBytes, U16 *u16reply)
{
	MPIDefaultReply_t *mpiReply;
	U8 *req_as_bytes;
	int r;
	int i;

	/*  Make sure there are no doorbells  */
	CHIPREG_WRITE32(&ioc->chip->IntStatus, 0);

	CHIPREG_WRITE32(&ioc->chip->Doorbell,
			((MPI_FUNCTION_HANDSHAKE<<MPI_DOORBELL_FUNCTION_SHIFT) |
			 ((reqBytes/4)<<MPI_DOORBELL_ADD_DWORDS_SHIFT)));

	/*  Wait for IOC doorbell int  */
	if ((i = WaitForDoorbellInt(ioc)) < 0)
		return i;

	dhsprintk((KERN_INFO "mpi: %s: HandShake request start, WaitCnt=%d\n", ioc->name, i));

	CHIPREG_WRITE32(&ioc->chip->IntStatus, 0);

	if ((r = WaitForDoorbellAck(ioc)) < 0)
		return -2;

	mpiReply = (MPIDefaultReply_t*) ioc->hs_reply;
	mpiReply->MsgLength = 0;
	ioc->hs_reply_idx = 0;

	/*  Send request via doorbell handshake  */
	req_as_bytes = (U8 *) req;
	for (i = 0; i < reqBytes/4; i++) {
		U32 word;

		word = ((req_as_bytes[(i*4) + 0] <<  0) |
			(req_as_bytes[(i*4) + 1] <<  8) |
			(req_as_bytes[(i*4) + 2] << 16) |
			(req_as_bytes[(i*4) + 3] << 24));
		CHIPREG_WRITE32(&ioc->chip->Doorbell, word);
		if ((r = WaitForDoorbellAck(ioc)) < 0)
			return -3;
	}

	/*  Wait for completion of doorbell reply from IOC  */
	if ((i = WaitForDoorbellReply(ioc)) < 0)
		return i;

	dhsprintk((KERN_INFO "mpi: %s: HandShake reply wait done, WaitCnt=%d\n", ioc->name, i));

	/*  Copy out the cached reply...  */
	for(i=0; i < MIN(replyBytes/2,mpiReply->MsgLength*2); i++)
		u16reply[i] = ioc->hs_reply[i];
	mpiReply->MsgLength = 0;
	ioc->hs_reply_idx = 0;

#ifdef MPI_DEBUG_HANDSHAKE
	{
		U32 *m = (U32 *) u16reply;

		mpiReply = (MPIDefaultReply_t*) u16reply;
		dhsprintk((KERN_INFO "mpi:HS %s handShake reply:\n", ioc->name));
		for (i = 0; i < mpiReply->MsgLength; i++)
			dhsprintk((KERN_INFO "  %08x%s\n",
				   cpu_to_le32(m[i]),
				   (i==3 ?
				    " (IOCStatus,Rsvd0xFFFF)" :
				    (i==4 ? " (IOCLogInfo)" : ""))));
	}
#endif

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	WaitForDoorbellAck - Wait for IOC to clear IOP_DOORBELL_STATUS bit
 *	@ioc: Pointer to MPI_ADAPTER structure
 *
 *	This routine waits for IOC doorbell handshake acknowledge.
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
WaitForDoorbellAck(MPI_ADAPTER *ioc)
{
	int count = 0;
	int cntdn = 10 * 100;
	U32 intstat;

	for(;cntdn;--cntdn,count++) {
		intstat = CHIPREG_READ32(&ioc->chip->IntStatus);
		if (! (intstat & MPI_HIS_IOP_DOORBELL_STATUS))
			break;

		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(1);

		if (!cntdn) {
			printk(KERN_ERR "mpi: ERROR: %s: Doorbell ACK timeout! (cnt=%d)\n", ioc->name, count);
			return -1;
		}
	}

	dhsprintk((KERN_INFO "mpi:HS WaitForDoorbell ACK (cnt=%d)\n", count));
	return count;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	WaitForDoorbellInt - Wait for IOC to set HIS_DOORBELL_INTERRUPT bit
 *	@ioc: Pointer to MPI_ADAPTER structure
 *
 *	This routine waits for IOC doorbell interrupt.
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
WaitForDoorbellInt(MPI_ADAPTER *ioc)
{
	int count = 0;
	int cntdn = 10 * 100;
	U32 intstat;

	while (--cntdn) {
		intstat = CHIPREG_READ32(&ioc->chip->IntStatus);
		if (intstat & MPI_HIS_DOORBELL_INTERRUPT) {
			break;
		}

		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(1);

		count++;
	}

	if (!cntdn) {
		printk(KERN_ERR "mpi: ERROR: %s: Doorbell INT timeout! (cnt=%d)\n", ioc->name, count);
		return -1;
	}

	dhsprintk((KERN_INFO "mpi:HS WaitForDoorbell INT (cnt=%d)\n", count));
	return count;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	WaitForDoorbellReply - Wait for and capture a IOC handshake reply
 *	@ioc: Pointer to MPI_ADAPTER structure
 *
 *	This routine polls the IOC for a handshake reply, 16 bits at a time.
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
WaitForDoorbellReply(MPI_ADAPTER *ioc)
{
	int u16cnt = 0;
	int fail = 0;
	int r;
	U16 *hs_reply = ioc->hs_reply;
 	volatile MPIDefaultReply_t *mpiReply = (MPIDefaultReply_t *) ioc->hs_reply;

	mpiReply->IOCStatus = cpu_to_le16(MPI_IOCSTATUS_SUCCESS);
	mpiReply->MsgLength = 0;

	/*  First U16 hunk  */
	if ((r = WaitForDoorbellInt(ioc)) < 0)
	        fail++;

	hs_reply[u16cnt++] =
	        le16_to_cpu(CHIPREG_READ32(&ioc->chip->Doorbell) & 0x0000FFFF);
	CHIPREG_WRITE32(&ioc->chip->IntStatus, 0);

	/*  Second U16 hunk  */
	if ((r = WaitForDoorbellInt(ioc)) < 0)
	        fail++;

	hs_reply[u16cnt++] =
	        le16_to_cpu(CHIPREG_READ32(&ioc->chip->Doorbell) & 0x0000FFFF);
	CHIPREG_WRITE32(&ioc->chip->IntStatus, 0);

	if (mpiReply->MsgLength /* && !fail && mpiReply->Function */) {
		/*  Ok, piece together a reply 16 bits at a time...  */
		/*  GRRRRR...  */

		for ( ; u16cnt < 2*mpiReply->MsgLength; u16cnt++) {
			if ((r = WaitForDoorbellInt(ioc)) < 0)
				fail++;
			hs_reply[u16cnt] =
				le16_to_cpu(CHIPREG_READ32(&ioc->chip->Doorbell) & 0x0000FFFF);
			CHIPREG_WRITE32(&ioc->chip->IntStatus, 0);
		}

		if ((r = WaitForDoorbellInt(ioc)) < 0)
			fail++;
		CHIPREG_WRITE32(&ioc->chip->IntStatus, 0);

	} else {
		printk(KERN_ERR "mpi: ERROR: That's odd - no handshake reply from %s\n", ioc->name);
	}

	dhsprintk((KERN_INFO "mpi:HS WaitForDoorbell REPLY (cnt=%d)\n", u16cnt/2));
	return u16cnt/2;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	GetLanConfigPages - Fetch LANConfig pages
 *	@ioc: Pointer to MPI_ADAPTER structure
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
GetLanConfigPages(MPI_ADAPTER *ioc)
{
	Config_t		 config_req;
	ConfigReply_t		 config_reply;
	LANPage0_t		*page0;
	dma_addr_t		 page0_dma;
	LANPage1_t		*page1;
	dma_addr_t		 page1_dma;
	int			 i;
	int			 req_sz;
	int			 reply_sz;

/* LANPage0 */
	/*  Immediate destination (reply area)...  */
	reply_sz = sizeof(config_reply);
	memset(&config_reply, 0, reply_sz);

	/*  Ultimate destination...  */
	page0 = &ioc->lan_cnfg_page0;
	reply_sz = sizeof(*page0);
	memset(page0, 0, reply_sz);

	/*  Request area (config_req on the stack right now!)  */
	req_sz = sizeof(config_req);
	memset(&config_req, 0, req_sz);
	config_req.Function = MPI_FUNCTION_CONFIG;
	config_req.Action = MPI_CONFIG_ACTION_PAGE_READ;
//	config_req.Header.PageVersion = 0;
//	config_req.Header.PageLength = 0;
	config_req.Header.PageNumber = 0;
	config_req.Header.PageType = MPI_CONFIG_PAGETYPE_LAN;
//	config_req.PageAddress = 0;
	config_req.PageBufferSGE.u.Simple.FlagsLength =
			( ((MPI_SGE_FLAGS_LAST_ELEMENT |
			    MPI_SGE_FLAGS_END_OF_BUFFER |
			    MPI_SGE_FLAGS_END_OF_LIST |
			    MPI_SGE_FLAGS_SIMPLE_ELEMENT |
			    MPI_SGE_FLAGS_SYSTEM_ADDRESS |
			    MPI_SGE_FLAGS_32_BIT_ADDRESSING |
			    MPI_SGE_FLAGS_32_BIT_CONTEXT) << MPI_SGE_FLAGS_SHIFT) |
			  (U32)sizeof(*page0) );
	page0_dma = pci_map_single(ioc->pcidev, page0, sizeof(*page0), PCI_DMA_FROMDEVICE);
	config_req.PageBufferSGE.u.Simple.u.Address32 = cpu_to_le32(page0_dma);

	i = HandShakeReqAndReply(ioc, req_sz, (U32*)&config_req,
				reply_sz, (U16*)&config_reply);
	pci_unmap_single(ioc->pcidev, page0_dma, sizeof(*page0), PCI_DMA_FROMDEVICE);
	if (i != 0)
		return i;

	/*  Now byte swap the necessary LANPage0 fields  */

/* LANPage1 */
	/*  Immediate destination (reply area)...  */
	reply_sz = sizeof(config_reply);
	memset(&config_reply, 0, reply_sz);

	/*  Ultimate destination...  */
	page1 = &ioc->lan_cnfg_page1;
	reply_sz = sizeof(*page1);
	memset(page1, 0, reply_sz);

	/*  Request area (config_req on the stack right now!)  */
	req_sz = sizeof(config_req);
	memset(&config_req, 0, req_sz);
	config_req.Function = MPI_FUNCTION_CONFIG;
	config_req.Action = MPI_CONFIG_ACTION_PAGE_READ;
//	config_req.Header.PageVersion = 0;
//	config_req.Header.PageLength = 0;
	config_req.Header.PageNumber = 1;
	config_req.Header.PageType = MPI_CONFIG_PAGETYPE_LAN;
//	config_req.PageAddress = 0;
	config_req.PageBufferSGE.u.Simple.FlagsLength =
			( ((MPI_SGE_FLAGS_LAST_ELEMENT |
			    MPI_SGE_FLAGS_END_OF_BUFFER |
			    MPI_SGE_FLAGS_END_OF_LIST |
			    MPI_SGE_FLAGS_SIMPLE_ELEMENT |
			    MPI_SGE_FLAGS_SYSTEM_ADDRESS |
			    MPI_SGE_FLAGS_32_BIT_ADDRESSING |
			    MPI_SGE_FLAGS_32_BIT_CONTEXT) << MPI_SGE_FLAGS_SHIFT) |
			  (U32)sizeof(*page1) );
	page1_dma = pci_map_single(ioc->pcidev, page1, sizeof(*page1), PCI_DMA_FROMDEVICE);
	config_req.PageBufferSGE.u.Simple.u.Address32 = cpu_to_le32(page1_dma);

	i = HandShakeReqAndReply(ioc, req_sz, (U32*)&config_req,
				reply_sz, (U16*)&config_reply);
	pci_unmap_single(ioc->pcidev, page1_dma, sizeof(*page1), PCI_DMA_FROMDEVICE);
	if (i != 0)
		return i;

	/*  Now byte swap the necessary LANPage1 fields  */

	return 0;
}

#ifdef CONFIG_PROC_FS
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	procfs (/proc/mpi/...) support stuff...
 */

#define PROC_MPI_READ_RETURN(page,start,off,count,eof,len) \
    {				\
        len -= off;		\
        if (len < count) {	\
            *eof = 1;		\
            if (len <= 0)	\
                return 0;	\
        } else			\
            len = count;	\
        *start = page + off;	\
        return len;		\
    }


/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	procmpi_create - Create /proc/mpi entries
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
procmpi_create(void)
{
	MPI_ADAPTER *ioc;
	struct proc_dir_entry *ent;

	procmpi_root_dir = CREATE_PROCDIR_ENTRY("mpi", 0);
	if (procmpi_root_dir == NULL)
		return -1;

	ioc = mpi_adapter_find_first();

	if (ioc) {
		ent = create_proc_read_entry("mpi/info", 0, NULL, procmpi_read_info, NULL);
		if (ent == NULL) {
			printk(KERN_ERR "mpi: ERROR: %s: Cannot create /proc/mpi/info entry!\n", ioc->name);
			remove_proc_entry("mpi", 0);
			return -ENOMEM;
		}
	}

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	procmpi_destroy - Tear down /proc/mpi entries
 *
 *	Returns 0 for success, non-zero for failure.
 */
STATIC int
procmpi_destroy(void)
{
	MPI_ADAPTER *ioc;

	ioc = mpi_adapter_find_first();
	while (ioc != NULL)
		ioc = mpi_adapter_find_next(ioc);

	remove_proc_entry("mpi/info", 0);

	if (! procmpi_root_dir->count)
		remove_proc_entry("mpi", 0);
	else
		return -1;

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	procmpi_read_info - Handle read request from /proc/mpi/info
 *	@page: Pointer to area to write information
 *	@start: Pointer to start pointer
 *	@off: Offset to start writing
 *	@count: 
 *	@eof: Pointer to EOF integer
 *	@data: Pointer 
 *
 *	Returns numbers of characters written to process performing the read.
 */
STATIC int
procmpi_read_info(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	MPI_ADAPTER *ioc;
	char *out = page;
	int len;

//	if (off > 0)
//		return 0;

	ioc = mpi_adapter_find_first();

	out += sprintf(out, "Attached Fusion MPT I/O Controllers:%s\n", ioc ? "" : " none");

	while (ioc) {
		int  more = 0;

		mpi_print_ioc_facts(ioc, out, &more, 0);
		out += more;
		if ((out-page) >= count) {
			break;
		}

		ioc = mpi_adapter_find_next(ioc);
	}
	len = out - page;

	PROC_MPI_READ_RETURN(page,start,off,count,eof,len);
}

#endif		/* CONFIG_PROC_FS */

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_print_ioc_facts - Write ASCII summary of IOC facts to a buffer
 *	@ioc: Pointer to MPI_ADAPTER structure
 *	@buffer: Pointer to buffer where IOC info should be written
 *	@size: Pointer to number of bytes we wrote (set by this routine)
 *	@len: Offset at which to start writing in buffer
 *
 * 	This routine writes (english readable) ASCII text, which represents
 * 	a summary of the IOC facts, to a buffer.
 */
void
mpi_print_ioc_facts(MPI_ADAPTER *ioc, char *buffer, int *size, int len)
{
	char iocName[16];
	char expVer[32];
	int y;
	int sz;

	strcpy(iocName, ioc->name);

	if ((ioc->facts0.FWVersion & 0xF000) == 0xE000)
		sprintf(expVer, "  (Exp %02d%02d)",
			(ioc->facts0.FWVersion & 0x0F00) >> 8,	/* Month */
			ioc->facts0.FWVersion & 0x001F);		/* Day   */
	else
		expVer[0] ='\0';

	/* insider hack! */
	if (ioc->facts0.FWVersion & 0x0080) {
		strcat(expVer, " [MDBG]");
	}

	y = sprintf(buffer+len, "%s:\n", iocName);

	y += sprintf(buffer+len+y, "  ProductID = 0x%04x\n", ioc->facts0.ProductID);
	y += sprintf(buffer+len+y, "  PortNumber = %d (of %d)\n",
		ioc->pfacts0.PortNumber+1,
		ioc->facts0.NumberOfPorts);
	if (ioc->pfacts0.ProtocolFlags & MPI_PORTFACTS_PROTOCOL_LAN) {
		U8 *a = (U8*)&ioc->lan_cnfg_page1.HardwareAddressLow;
		y += sprintf(buffer+len+y, "  LanAddr = 0x(%02x:%02x:%02x:%02x:%02x:%02x)\n",
			a[5], a[4], a[3], a[2], a[1], a[0]);
	}
	y += sprintf(buffer+len+y, "  FWVersion = 0x%04x%s\n", ioc->facts0.FWVersion, expVer);
	y += sprintf(buffer+len+y, "  MsgVersion = 0x%04x\n", ioc->facts0.MsgVersion);
	y += sprintf(buffer+len+y, "  WhoInit = 0x%02x\n", ioc->facts0.WhoInit);
	y += sprintf(buffer+len+y, "  EventState = 0x%02x\n", ioc->facts0.EventState);
	y += sprintf(buffer+len+y, "  CurrentHostMfaHighAddr = 0x%08x\n",
		     ioc->facts0.CurrentHostMfaHighAddr);
	y += sprintf(buffer+len+y, "  CurrentSenseBufferHighAddr = 0x%08x\n",
		     ioc->facts0.CurrentSenseBufferHighAddr);
	y += sprintf(buffer+len+y, "  MaxChainDepth = 0x%02x frames\n", ioc->facts0.MaxChainDepth);
	y += sprintf(buffer+len+y, "  MinBlockSize = 0x%02x bytes\n", 4*ioc->facts0.BlockSize);

	y += sprintf(buffer+len+y, "  RequestFrames @ 0x%p (Dma @ 0x%08x)\n",
					ioc->req_alloc, ioc->req_alloc_dma);
	/*
	 *  Rounding UP to nearest 4-kB boundary here...
	 */
	sz = (ioc->req_sz * ioc->req_depth) + 128;
	sz = ((sz + 0x1000UL - 1UL) / 0x1000) * 0x1000;
	y += sprintf(buffer+len+y, "    {CurReqSz=%d} x {CurReqDepth=%d} = %d bytes ^= 0x%x\n",
					ioc->req_sz, ioc->req_depth, ioc->req_sz*ioc->req_depth, sz);
	y += sprintf(buffer+len+y, "    {MaxReqSz=%d}   {MaxReqDepth=%d}\n",
					4*ioc->facts0.RequestFrameSize,
					ioc->facts0.GlobalCredits);

	y += sprintf(buffer+len+y, "  ReplyFrames   @ 0x%p (Dma @ 0x%08x)\n",
					ioc->reply_alloc, ioc->reply_alloc_dma);
	sz = (ioc->reply_sz * ioc->reply_depth) + 128;
	y += sprintf(buffer+len+y, "    {CurRepSz=%d} x {CurRepDepth=%d} = %d bytes ^= 0x%x\n",
					ioc->reply_sz, ioc->reply_depth, ioc->reply_sz*ioc->reply_depth, sz);
	y += sprintf(buffer+len+y, "    {MaxRepSz=%d}   {MaxRepDepth=%d}\n",
					ioc->factsN.CurReplyFrameSize,
					ioc->facts0.ReplyQueueDepth);

	*size = y;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#ifdef MODULE
EXPORT_SYMBOL(mpi_register);
EXPORT_SYMBOL(mpi_deregister);
EXPORT_SYMBOL(mpi_register_ascqops_strings);
EXPORT_SYMBOL(mpi_deregister_ascqops_strings);
EXPORT_SYMBOL(ASCQ_TablePtr);
EXPORT_SYMBOL(ASCQ_TableSz);
EXPORT_SYMBOL(ScsiOpcodesPtr);
EXPORT_SYMBOL(MpiGetMsgFrame);
EXPORT_SYMBOL(MpiPutMsgFrame);
EXPORT_SYMBOL(MpiFreeMsgFrame);
EXPORT_SYMBOL(SendHandShakeRequest);
EXPORT_SYMBOL(mpi_adapter_find_first);
EXPORT_SYMBOL(mpi_adapter_find_next);
EXPORT_SYMBOL(mpi_verify_adapter);
EXPORT_SYMBOL(mpi_print_ioc_facts);
EXPORT_SYMBOL(LanCIdx);
#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	init_module - Fusion MPT base driver initialization routine
 *
 *	Returns 0 for success, non-zero for failure.
 */
int
#ifdef MODULE
init_module(void)
#else
my_init(void)
#endif
{
	int i;

	show_mpimod_ver(my_NAME, my_VERSION);

	Q_INIT(&MpiAdapters, MPI_ADAPTER);			/* set to empty */
	for (i = 0; i < 8; i++)
		MpiCallbacks[i] = NULL;

	if ((i = mpi_pci_scan()) < 0)
		return i;

	return 0;
}

#ifdef MODULE
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	cleanup_module - Perform module unload cleanup
 *
 *	This routine frees all resources associated with each MPI adapter
 *	and removes all /proc/mpi entries.
 */
void
cleanup_module(void)
{
	MPI_ADAPTER *this;

	while (! Q_IS_EMPTY(&MpiAdapters)) {
		this = MpiAdapters.head;
		Q_DEL_ITEM(this);
		mpi_adapter_dispose(this);
	}

#ifdef CONFIG_PROC_FS
	procmpi_destroy();
#endif
}

#else

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_init - Fusion MPT base driver initialization routine
 *	when driver is compiled into the kernel (not as a module)
 */
void __init
mpi_init(void)
{
	my_init();
}

#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

